/*****************************************************************************
*
*
* File              : RS_232_Logger_Master.c
*
*   Aufgezeichnung und Darstellung zweier RS232 Zeichenstroeme.
*   Es werden zwei Kanaele quasiparallel aufgezeichnet und mit
*   einem Zeitstempel, mit einer Millisekunde Aufloesung, in
*   einen gemeinsamen Puffer geschrieben. Anschlieend bereitet
*   die Hauptschleife die Informationen auf.
*
*   (c) 2013 ths
*
*   1.04 - Neu auf M32 statt M8515 mehr Puffer
*          Achtung im neuen Layout Port A unf C getauscht bezglicht
*          INT_0 und INT_1   
*
****************************************************************************/

#include <ctype.h>
#include <avr/io.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>



const char  COPYR[]    PROGMEM = { "\n\r\n\r"
                                   "*** Mega-32-RS232-Logger  "
                                   __DATE__
                                   "  V1.04"
                                   " (c) ths ***"
                                   "\n\r\n\r" };

const char  OK[]       PROGMEM = "\n\rok.\n\r";

const char  HEX_STR[]  PROGMEM = "0123456789ABCDEF";

#define  BELL       7
#define  LF         10
#define  CR         13
#define  ESC        27

#define  CX_1       0x11
#define  CX_2       0x22

#define  LED_HB     PD6
#define  RESET_FR   PD7

#define  ACK_1      PB0
#define  ACK_2      PB1

#define  STROBE_1   PD2
#define  STROBE_2   PD3


// EEPROM Variablen
//
unsigned char  eeDummy              EEMEM;     // Dummy damit Adresse 0 nicht benutzt wird
unsigned char  eeMagic              EEMEM;     // Magic Byte
unsigned char  eeBaudFr             EEMEM;     // Baudrate Frontends
unsigned char  eeLengthFr           EEMEM;     // Zeichenlaenge Frontends
unsigned char  eeModeAnz            EEMEM;     // Anzeige Modus Alternierend / Singletimestamp
unsigned char  eeModeBase           EEMEM;     // Anzeige Zahlenformat ASCII / HEX
unsigned char  eeLineLength         EEMEM;     // Anzeige Zeilenlaenge


//--------------------------------------------------------------------------
// PC V24
//
#define  BAUDRATE        19200
#define  MAX_BUF_V24     16

volatile char            buffer[ MAX_BUF_V24 ];
volatile char *          ip_V24;
volatile char *          op_V24;

//--------------------------------------------------------------------------
// Logging Buffer
//
#define  MAX_BUF         448

typedef  struct { unsigned int   time;
                  char           chn;
                  unsigned char  data;
                } sbuffer;

volatile sbuffer       puffer[ MAX_BUF ];
volatile unsigned int  ip_buf;
volatile unsigned int  op_buf;
volatile unsigned int  tics;

//--------------------------------------------------------------------------
//
#define  ALT             11
#define  SINGLE          12
#define  ASCII           13
#define  HEX             14

#define  SPACE           15
#define  NO_SPACE        16
   
volatile char            BaudFr     = 10;      // 19200 
volatile char            LengthFr   = 0x30;    // 8 Bit
volatile char            ModeAnz    = ALT;   
volatile char            ModeBase   = HEX;
volatile char            LineLength = 80;

/*----------------------------------------------------------------------------*/
void io_putch( unsigned char );
void io_space( void );
void io_crlf( void );
void io_phex_u8( unsigned char );
void io_phex_u16( unsigned int );
void io_pdez_u8( unsigned char );
void io_pdez_u16( unsigned int );
void io_pstr( const char * );
char io_getch( void );
char io_kbhit( void );

void p_baud( void );
void p_bit_len( void );
void p_help( void );
int  p_data( unsigned char, char );
int  p_time( char, unsigned int );
int  p_time_data( char, unsigned int, unsigned char );

void show_baud( void );
void show_bit_len( void );
void show_status( void );
void reset_frontend( unsigned char );

/*----------------------------------------------------------------------------*/
// Eine Millisekunde Timer Intervall
//
ISR( TIMER1_COMPA_vect )
/*=====================*/
{
  tics++;
}

/*----------------------------------------------------------------------------*/
ISR( INT0_vect )
/*=============*/
{
  unsigned char  tmp;

  // Hole Datum vom Port 1
  //
  tmp = PINC;

  // ACK_1 auf low
  //
  PORTB &= ~_BV( ACK_1 );

  // Schreibe Datum in Puffer
  //
  puffer[ ip_buf ].time = tics;
  puffer[ ip_buf ].chn  = CX_1;
  puffer[ ip_buf ].data = tmp;

  ip_buf++;
  if ( ip_buf >= MAX_BUF )
  {
    ip_buf = 0;
  }

  // Warte auf Strobe == high
  //
  loop_until_bit_is_set( PIND, PD2 ) ;

  // ACK_1 auf high
  //
  PORTB |= _BV( ACK_1 );
}

/*----------------------------------------------------------------------------*/
ISR( INT1_vect )
/*=============*/
{
  unsigned char  tmp;

  // Hole Datum vom Port 2
  //
  tmp = PINA;

  // ACK_2 auf low
  //
  PORTB &= ~_BV( ACK_2 );

  // Schreibe Datum in Puffer
  //
  puffer[ ip_buf ].time = tics;
  puffer[ ip_buf ].chn  = CX_2;
  puffer[ ip_buf ].data = tmp;

  ip_buf++;
  if ( ip_buf >= MAX_BUF )
  {
    ip_buf = 0;
  }

  // Warte auf Strobe == high
  //
  loop_until_bit_is_set( PIND, PD3 ) ;

  // ACK_2 auf high
  //
  PORTB |= _BV( ACK_2 );
}

/*----------------------------------------------------------------------------*/
ISR( USART_RXC_vect )
/*==================*/
{
  *ip_V24++ = UDR;

  if ( ip_V24 >= &( buffer[ MAX_BUF_V24 ] ) )
  {
    ip_V24  = &( buffer[ 0 ] );
  }
}

/*============================================================================*/
void io_putch( unsigned char  ch )
/*===============================*/
{
  loop_until_bit_is_set( UCSRA, UDRE ) ;

  UDR = ch;
}

/*----------------------------------------------------------------------------*/
void io_space( void )
/*==================*/
{
  io_putch( ' ' );
}

/*----------------------------------------------------------------------------*/
void io_crlf( void )
/*=================*/
{
  io_putch( '\n' );
  io_putch( '\r' );
}

/*-------------------------------------------------------------------------*/
/* Ausgabe eines Hexwertes von 0..FF                                       */
/*-------------------------------------------------------------------------*/
void io_phex_u8( unsigned char  val )
/*==================================*/
{
  const char *  p;

  p = (const char *)&( HEX_STR[ val / 16 ] );
  io_putch( pgm_read_byte( p ) );

  p = (const char *)&( HEX_STR[ val % 16 ] );
  io_putch( pgm_read_byte( p ) );
}

/*-------------------------------------------------------------------------*/
/* Ausgabe eines Hexwertes von 0..FFFF                                     */
/*-------------------------------------------------------------------------*/
void io_phex_u16( unsigned int  val )
/*==================================*/
{
  io_phex_u8( val >> 8 );
  io_phex_u8( val & 0xFF );
}

/*-------------------------------------------------------------------------*/
/* Ausgabe eines Dezimalwertes von 0..255                                  */
/*-------------------------------------------------------------------------*/
void io_pdez_u8( unsigned char  val )
/*==================================*/
{
  unsigned char   digit = 0;

  while ( val >= 100 )
  {
    val -= 100;
    digit++;
  }
  io_putch( digit + '0' );

  digit = 0;
  while ( val >= 10 )
  {
    val -= 10;
    digit++;
  }
  io_putch( digit + '0' );
  io_putch( val   + '0' );
}

/*-------------------------------------------------------------------------*/
/* Ausgabe eines Dezimalwertes von 0..65535                                */
/*-------------------------------------------------------------------------*/
void io_pdez_u16( unsigned int  val )
/*==================================*/
{
  unsigned char   digit = 0;

  while ( val >= 10000 )
  {
    val -= 10000;
    digit++;
  }
  io_putch( digit + '0' );

  digit = 0;
  while ( val >= 1000 )
  {
    val -= 1000;
    digit++;
  }
  io_putch( digit + '0' );

  digit = 0;
  while ( val >= 100 )
  {
    val -= 100;
    digit++;
  }
  io_putch( digit + '0' );

  digit = 0;
  while ( val >= 10 )
  {
    val -= 10;
    digit++;
  }
  io_putch( digit + '0' );
  io_putch( val   + '0' );
}

/*----------------------------------------------------------------------------*/
void io_pstr( const char * p )
/*===========================*/
{
  register char  c;

  while ( ( c = pgm_read_byte( p++ ) ) )
  {
    io_putch( c );
  }
}

/*----------------------------------------------------------------------------*/
char  io_getch( void )
/*===================*/
{
  char  ch;

  // mu nicht ATOMIC sein, da nur ein Byte Pointer
  //
  if ( ip_V24 == op_V24 )
  {
    return( 0 );
  }

  ch = *op_V24++;
  if ( op_V24 >= &( buffer[ MAX_BUF_V24 ] ) )
  {
    op_V24  = &( buffer[ 0 ] );
  }

  return( ch );
}

/*----------------------------------------------------------------------------*/
char  io_kbhit( void )
/*===================*/
{
  if ( ip_V24 != op_V24 )
  {
    return( 1 );
  }
  else
  {
    return( 0 );
  }
}

/*----------------------------------------------------------------------------*/
void p_baud( void )
/*================*/
{
  io_crlf( );
  io_pstr( PSTR( "0 =   75           8 =   9600" ) );
  io_crlf( );
  io_pstr( PSTR( "1 =  150           9 =  14400" ) );
  io_crlf( );
  io_pstr( PSTR( "2 =  300           A =  19200" ) );
  io_crlf( );
  io_pstr( PSTR( "3 =  600           B =  28800" ) );
  io_crlf( );
  io_pstr( PSTR( "4 = 1200           C =  38400" ) );
  io_crlf( );
  io_pstr( PSTR( "5 = 2400           D =  57600" ) );
  io_crlf( );
  io_pstr( PSTR( "6 = 4800           E =  76800" ) );
  io_crlf( );
  io_pstr( PSTR( "7 = 7200           F = 115200" ) );
  io_crlf( );
  io_crlf( );
}

/*----------------------------------------------------------------------------*/
void p_bit_len( void )
/*===================*/
{
  io_crlf( );
  io_pstr( PSTR( "0 =   5 Bit" ) );
  io_crlf( );
  io_pstr( PSTR( "1 =   6 Bit" ) );
  io_crlf( );
  io_pstr( PSTR( "2 =   7 Bit" ) );
  io_crlf( );
  io_pstr( PSTR( "3 =   8 Bit" ) );
  io_crlf( );
  io_crlf( );
}

/*----------------------------------------------------------------------------*/
void p_help( void )
/*================*/
{
  io_crlf( );
  io_crlf( );
  io_pstr( PSTR( "   T - Altenate Timestamp Mode <00001> t e s t 0D ..." ) );
  io_crlf( );
  io_pstr( PSTR( "   S - Single Timestamp Mode   <00001:t><00002:e>...." ) );
  io_crlf( );
  io_crlf( );
  io_pstr( PSTR( "   A - ASCII Mode" ) );
  io_crlf( );
  io_pstr( PSTR( "   H - HEX Mode" ) );
  io_crlf( );
  io_crlf( );
  io_pstr( PSTR( "   M - Mode ausgeben        <> = Kanal 1" ) );
  io_crlf( );
  io_pstr( PSTR( "   C - Clear Timer          [] = Kanal 2" ) );
  io_crlf( );
  io_crlf( );
  io_pstr( PSTR( "   B - Baudrate Frontend" ) );
  io_crlf( );
  io_pstr( PSTR( "   W - Zeichenlaenge Frontend" ) );
  io_crlf( );
  io_pstr( PSTR( "   X - Reset Frontend ( Tiny2313 )" ) );
  io_crlf( );
  io_crlf( );
}

/*----------------------------------------------------------------------------*/
int  p_data( unsigned char  data, char  flag )
/*===========================================*/
{
  int  len = 0;

  if ( LengthFr == 0x00 )
  {
    data &= 0x1F;
  }
  else if ( LengthFr == 0x10 )
  {
    data &= 0x3F;
  }
  else if ( LengthFr == 0x20 )
  {
    data &= 0x7F;
  }
  else
  {
    // bleibt so
    //
  }

  if ( ModeBase == ASCII )
  {
    // Zeichen druckbar ??
    //
    if ( ( data > 0x20 ) && ( data <= 0x7F ) )
    {
      // Wenn ja, als ASCII ausgeben
      //
      io_putch( data );
      len = 1;
    }
    else
    {
      // Wenn nicht, als HEX-Byte ausgeben
      //
      io_phex_u8( data );
      len = 2;
    }
  } 
  else if ( ModeBase == HEX )
  {
    // Alles in HEX ausgeben
    //
    io_phex_u8( data );
    len = 2;
  }
  else
  {
    io_pstr( PSTR( "Bad Base Mode\n\r" ) );
  }

  if ( flag == SPACE )
  {
    io_putch( ' ' );
    len++;
  }

  return( len );
}

/*----------------------------------------------------------------------------*/
int  p_time( char  chn, unsigned int  time )
/*=========================================*/
{
  int  len = 0;

  if ( chn == CX_1 )
  {
    io_putch( '<' );
    io_pdez_u16( time );
    io_putch( '>' );
    io_putch( ' ' );
    len = 8;
  }
  else if ( chn == CX_2 )
  {
    io_putch( '[' );
    io_pdez_u16( time );
    io_putch( ']' );
    io_putch( ' ' );
    len = 8;
  }
  else
  {
    io_pstr( PSTR( "Bad CHN\n\r" ) );
  }

  return( len );
}

/*----------------------------------------------------------------------------*/
int  p_time_data( char  chn, unsigned int  time, unsigned char  data )
/*===================================================================*/
{
  int  len = 0;

  if ( chn == CX_1 )
  {
    io_putch( '<' );
    io_pdez_u16( time );
    io_putch( ':' );
    len = 7;
      
    len += p_data( data, NO_SPACE );
      
    io_putch( '>' );
    len++;
  }
  else if ( chn == CX_2 )
  {
    io_putch( '[' );
    io_pdez_u16( time );
    io_putch( ':' );
    len = 7;
      
    len += p_data( data, NO_SPACE );
      
    io_putch( ']' );
    len++;
  }
  else
  {
    io_pstr( PSTR( "Bad CHN\n\r" ) );
  }

  return( len );
}

/*----------------------------------------------------------------------------*/
void show_baud( void )
/*===================*/
{
  switch ( BaudFr & 0x0F )
  {
    case  0:  io_pstr( PSTR(     "75" ) );  break;
    case  1:  io_pstr( PSTR(    "150" ) );  break;
    case  2:  io_pstr( PSTR(    "300" ) );  break;
    case  3:  io_pstr( PSTR(    "600" ) );  break;
    case  4:  io_pstr( PSTR(   "1200" ) );  break;
    case  5:  io_pstr( PSTR(   "2400" ) );  break;
    case  6:  io_pstr( PSTR(   "4800" ) );  break;
    case  7:  io_pstr( PSTR(   "7200" ) );  break;
    case  8:  io_pstr( PSTR(   "9600" ) );  break;
    case  9:  io_pstr( PSTR(  "14400" ) );  break;
    case 10:  io_pstr( PSTR(  "19200" ) );  break;
    case 11:  io_pstr( PSTR(  "28800" ) );  break;
    case 12:  io_pstr( PSTR(  "38400" ) );  break;
    case 13:  io_pstr( PSTR(  "57600" ) );  break;
    case 14:  io_pstr( PSTR(  "76800" ) );  break;
    case 15:  io_pstr( PSTR( "115200" ) );  break;
  }

  io_pstr( PSTR( " Bd" ) );
}

/*----------------------------------------------------------------------------*/
void show_bit_len( void )
/*===================*/
{
  switch ( LengthFr & 0x30 )
  {
    case 0x00:  io_putch( '5' );  break;
    case 0x10:  io_putch( '6' );  break;
    case 0x20:  io_putch( '7' );  break;
    case 0x30:  io_putch( '8' );  break;
  }

  io_pstr( PSTR( " Bit" ) );
}


/*----------------------------------------------------------------------------*/
void show_status( void )
/*=====================*/
{
  io_crlf( );
  io_pstr( PSTR( "Status = " ) );
  if ( ModeAnz == SINGLE )
  {
    io_pstr( PSTR( "Single Timestamp" ) );
  }
  else if ( ModeAnz == ALT )
  {
    io_pstr( PSTR( "Alternate Timestamp" ) );
  }
  else
  {
    io_pstr( PSTR( "illegal" ) );
  }

  io_pstr( PSTR( ",  " ) );

  if ( ModeBase == ASCII )
  {
    io_pstr( PSTR( "ASCII ( 0x1F < CHR < 0x7F )" ) );
  }
  else if ( ModeBase == HEX )
  {
    io_pstr( PSTR( "HEX ( 0x00 .. 0xFF )" ) );
  }
  else
  {
    io_pstr( PSTR( "illegal" ) );
  }
      
  io_pstr( PSTR( ",   " ) );
  show_baud( );

  io_pstr( PSTR( ",  " ) );
  show_bit_len( );
   
  io_crlf( );
}

/*----------------------------------------------------------------------------*/
void reset_frontend( unsigned char  config )
/*=========================================*/
{
  // Reset ziehen
  //
  io_crlf( );
  io_pstr( PSTR( "Config = " ) );
  io_phex_u8( config );
  io_crlf( );

  PORTD &= ~_BV( RESET_FR );
  _delay_ms( 3 );
  PORTD |=  _BV( RESET_FR );

  // Datenrichtung auf Ausgang
  //
  DDRA  = 0b11111111;
  PORTA = config;

  DDRC  = 0b11111111;
  PORTC = config;

  // Reset + 64ms beim Slave
  //
  _delay_ms( 80 );

  // wieder auf Normalbetrieb schalten
  //
  // Komunikationsport 1
  // PA0..7 = Daten vom Logger
  // PB0    = ACK zum Logger
  // PD2    = IRQ vom Logger
  //
  DDRA  = 0b00000000;
  PORTA = 0b00000000;

  // Komunikationsport 2
  // PC0..7 = Daten vom Logger
  // PB1    = ACK zum Logger
  // PD3    = IRQ vom Logger
  //
  DDRC  = 0b00000000;
  PORTC = 0b00000000;
}

/*----------------------------------------------------------------------------*/
void set_baud( void )
/*==================*/
{
  char           done;
  char           key;

  p_baud( );   
  io_pstr( PSTR( "Waehle Baudrate [ 0..F ]   > " ) );
   
  done = 0;
  do
  {
    if ( io_kbhit( ) == 1 )
    {
      key = io_getch( );
      io_putch( key );

      switch ( key )
      {
        case '0':  BaudFr = 0x00;  done = 1;  break;
        case '1':  BaudFr = 0x01;  done = 1;  break;
        case '2':  BaudFr = 0x02;  done = 1;  break;
        case '3':  BaudFr = 0x03;  done = 1;  break;
        case '4':  BaudFr = 0x04;  done = 1;  break;
        case '5':  BaudFr = 0x05;  done = 1;  break;
        case '6':  BaudFr = 0x06;  done = 1;  break;
        case '7':  BaudFr = 0x07;  done = 1;  break;
        case '8':  BaudFr = 0x08;  done = 1;  break;
        case '9':  BaudFr = 0x09;  done = 1;  break;
        case 'A':  BaudFr = 0x0A;  done = 1;  break;
        case 'a':  BaudFr = 0x0A;  done = 1;  break;
        case 'B':  BaudFr = 0x0B;  done = 1;  break;
        case 'b':  BaudFr = 0x0B;  done = 1;  break;
        case 'C':  BaudFr = 0x0C;  done = 1;  break;
        case 'c':  BaudFr = 0x0C;  done = 1;  break;
        case 'D':  BaudFr = 0x0D;  done = 1;  break;
        case 'd':  BaudFr = 0x0D;  done = 1;  break;
        case 'E':  BaudFr = 0x0E;  done = 1;  break;
        case 'e':  BaudFr = 0x0E;  done = 1;  break;
        case 'F':  BaudFr = 0x0F;  done = 1;  break;
        case 'f':  BaudFr = 0x0F;  done = 1;  break;

        case ESC:  done = 1;                  break;
      }

      if ( done == 1 )
      {
        break;
      }
    }
    else
    {
      PORTD ^= _BV( LED_HB );
      _delay_ms( 100 ); 
    }
  }
  while ( done == 0 ) ;

  io_pstr( OK );
}

/*----------------------------------------------------------------------------*/
void set_wlen( void )
/*==================*/
{
  char           done;
  char           key;

  p_bit_len( );
  io_pstr( PSTR( "Waehle Bitlaenge [ 0..3 ]   > " ) );

  done = 0;
  do
  {
    if ( io_kbhit( ) == 1 )
    {
      key = io_getch( );
      io_putch( key );

      switch ( key )
      {
         case '0':  LengthFr = 0x00;  done = 1;  break;
         case '1':  LengthFr = 0x10;  done = 1;  break;
         case '2':  LengthFr = 0x20;  done = 1;  break;
         case '3':  LengthFr = 0x30;  done = 1;  break;

         case ESC:  done = 1;                    break;
      }
    }
    else
    {
      PORTD ^= _BV( LED_HB );
      _delay_ms( 100 ); 
    }
  }
  while ( done == 0 );

  io_pstr( OK );
}

/*----------------------------------------------------------------------------*/
int  main( void )
/*==============*/
{
  long           beatcnt = 0;
  char           key;
  unsigned int   time;
  char           chn;
  unsigned char  data;
  int            line_len;
  char           last_chn;
  unsigned char  config;
  int            temp;

  // Komunikationsport 1
  // PA0..7 = Daten vom Logger
  // PB0    = ACK zum Logger
  // PD2    = IRQ vom Logger
  //
  DDRA  = 0b00000000;
  PORTA = 0b00000000;

  // Komunikationsport 2
  // PC0..7 = Daten vom Logger
  // PB1    = ACK zum Logger
  // PD3    = IRQ vom Logger
  //
  DDRC  = 0b00000000;
  PORTC = 0b00000000;

  //  s.o.
  //
  DDRB  = 0b11111111;
  PORTB = 0b00000011;

  // V24             PD0, 1
  // Externe IRQ's   PD2, 3
  // LED             PD6
  // Reset Frontend  PD7
  //
  DDRD  = 0b11000010;
  PORTD = 0b11111101;

  // Timer_1 auf 1 ms Intervall
  // 11059200 Hz / 11059 = 1 KHz = 1 ms
  //
  TCCR1A = 0b00000000;         // CTC
  TCCR1B = 0b00001001;         // div_1
  TCNT1  = 0x0000;
  OCR1A  = ( 11059 - 1 );
  TIMSK  = _BV( OCIE1A );

  // Init UART, Enable IRQ_RX, RxD und TxD
  // Baudrate auf 19200
  //
  UCSRA = 0b00000000;
  UCSRB = 0b10011000;

  UBRRH = (unsigned char)( ( ( F_CPU / ( BAUDRATE * 16L ) ) - 1 ) >> 8 );
  UBRRL = (unsigned char)(   ( F_CPU / ( BAUDRATE * 16L ) ) - 1 );

  // Setze Pointer fr V24-Buffer
  //
  ip_V24 = &( buffer[ 0 ] );
  op_V24 = &( buffer[ 0 ] );

  io_pstr( COPYR );

  // Magic_Byte aus EEPROM holen
  //
  if ( eeprom_read_byte( &eeMagic ) == 0x55 )
  {
    // Parameter aus EEPROM lesen
    //
    BaudFr     = eeprom_read_byte( &eeBaudFr     );
    LengthFr   = eeprom_read_byte( &eeLengthFr   );
    ModeAnz    = eeprom_read_byte( &eeModeAnz    );
    ModeBase   = eeprom_read_byte( &eeModeBase   );
    LineLength = eeprom_read_byte( &eeLineLength );
  }
  else
  {
    // Default ins EEPROM schreiben
    //
    eeprom_write_byte( &eeMagic,      0x55       );
    eeprom_write_byte( &eeBaudFr,     BaudFr     );
    eeprom_write_byte( &eeLengthFr,   LengthFr   );
    eeprom_write_byte( &eeModeAnz,    ModeAnz    );
    eeprom_write_byte( &eeModeBase,   ModeBase   );
    eeprom_write_byte( &eeLineLength, LineLength );
  }
      
  config =  BaudFr   & 0x0F;
  config |= LengthFr & 0x30;
  reset_frontend( config );

  p_help( );
  show_status( );

  ip_buf = 0;
  op_buf = 0;

  // Aktiviere IRQ bei fallender Flanke an INT_0 und INT_1
  //
  MCUCR = 0b00001010;
  GICR  = _BV( INT1 ) | _BV( INT0 );

  // ab jetzt geht es rund
  //
  sei( );

  // FOERVER-Loop
  //
  line_len = 0;
  last_chn = 0x55;

  for ( ; ; )
  {
    //---------------------------------------------
    // Ist was im Empfangs-Puffer
    //
	ATOMIC_BLOCK( ATOMIC_RESTORESTATE )
	{
	  temp = ip_buf;
	}
      
	if ( temp != op_buf )
    {
      time = puffer[ op_buf ].time;
      chn  = puffer[ op_buf ].chn;
      data = puffer[ op_buf ].data;

      op_buf++;
      if ( op_buf >= MAX_BUF )
      {
        op_buf = 0;
      }

      // Je nach Mode Zeichen formatiert ausgeben
      //
      if ( ModeAnz == ALT )
      {
         // im Format <00000> a b c 
         //           [00012] x y z 
         //
         if ( chn != last_chn )
         {
           last_chn = chn;
           line_len = 0;
           io_crlf( );

           // Bei Wechsel einmal den Zeitstempel und Kanal ausgeben
           //
           line_len += p_time( chn, time );
           line_len += p_data( data, SPACE );
         }
         else
         {
           line_len += p_data( data, SPACE );
         }
       }
       else if ( ModeAnz == SINGLE )
       {
         // im Format <00000:a>[00000:b]
         //
         line_len += p_time_data( chn, time, data );
       }
       else
       {
         io_pstr( PSTR( "Bad Anz Mode\n\r" ) );
         line_len = 0;
       }

       // Wenn die Zeilenlaenge ueberschritten wird
       // eine neue Zeile beginnen
       //
       if ( line_len > LineLength )
       {
         line_len = 0;
         io_crlf( );
       }
     }

     //---------------------------------------------
     // Ist eine Taste gedrckt worden,
     // dann will wohl wer was
     //
     if ( io_kbhit( ) == 1 )
     {
       key = io_getch( );
       line_len = 0;

       // Befehl auswerten und ausfuehren
       //
       switch ( key )
       {
         case 'M':
         case 'm':
           show_status( );
           break;

         case 'B':
         case 'b':
           set_baud( );
           config =  BaudFr   & 0x0F;
           config |= LengthFr & 0x30;
           reset_frontend( config );
           show_status( );
           eeprom_write_byte( &eeBaudFr, BaudFr );
           key = 0;
           break;

         case 'W':
         case 'w':
           set_wlen( );
           config =  BaudFr   & 0x0F;
           config |= LengthFr & 0x30;
           reset_frontend( config );
           show_status( );
           eeprom_write_byte( &eeLengthFr, LengthFr );
           break;

         case 'C':
         case 'c':
           // Atomar den Millisekundenzaehler auf Null stellen
           //
           ATOMIC_BLOCK( ATOMIC_RESTORESTATE )
           {
             tics = 0;
           }
           io_crlf( );
           break;

         case 'A':
         case 'a':
           ModeBase = ASCII;
           eeprom_write_byte( &eeModeBase, ModeBase );
           show_status( );
           break;

         case 'H':
         case 'h':
           ModeBase = HEX;
           eeprom_write_byte( &eeModeBase, ModeBase );
           show_status( );
           break;

         case 'S':
         case 's':
           ModeAnz = SINGLE;
           eeprom_write_byte( &eeModeAnz, ModeAnz );
           show_status( );
           break;

         case 'T':
         case 't':
           ModeAnz = ALT;
           eeprom_write_byte( &eeModeAnz, ModeAnz );
           show_status( );
           break;

         case 'X':
         case 'x':
           reset_frontend( config );
           io_pstr( PSTR( "  Reset Frontend ok" ) );
           io_crlf( );
           break;

         case 'V':
         case 'v':
           io_pstr( COPYR );
           break;

         case ' ':
           io_crlf( );
           io_pstr( PSTR( "--------------------------------------------" ) );
           io_crlf( );
           break;

         case '?':
         default:
           p_help( );
           break;

       } // switch ( key )

     } // if ( io_kbhit( ) == 1 )

     //---------------------------------------------
     // LED blinken lassen
     //
     beatcnt++;
     if ( beatcnt > 110230 )
     {
       beatcnt = 0;
       PORTD ^= _BV( LED_HB );
     }

  } // Forever

}


